#include <cpctelera.h>
#include "collider.h"
#include "scene.h"
#include "renderer.h"

#define ABS(x)  (x < 0? -x: x)

#define FX_SCALE        3
#define FY_SCALE        3

#define COLLIDER_MAX_GAP  4

void collider_create(Collider* collider) {
  collider->path = scene_paths;
  collider->platform = scene_platforms;
  collider->block = blocks;
  collider->onPath = false;
  collider->onBlock = false;
}

void collider_checkWithBlocks(Collider* collider, u16 x0, u8 y0, u16 x1, u8 y1) {
  Block*  block = collider->block;
  u16     x = (x0 + x1) >> 1;
  u8      yMin, yMax;

  // Update current block pointer
  while (block->x1 <= x) block = ++collider->block;
  while (block->x0 > x) block = --collider->block;

  if (y0 < y1) {
    yMin = y0; yMax = y1;
  } else {
    yMin = y1; yMax = y0;
  }

  if (block->x0 <= x && block->x1 > x && block->y1 > yMin && yMax >= block->y0) {
    collider->onFloor = block->climbable && block->y0 >= yMin && block->y0 <= yMax;
    collider->onBlock = true;
    collider->fx = INT2FIXED(x, FX_SCALE);
    collider->fy = INT2FIXED(block->y0, FY_SCALE);
    return;
  }
  collider->onBlock = false;
}

void collider_checkWithPaths(Collider* collider, Path** pPath, u16 x0, u8 y0, u16 x1, u8 y1) {
  Path* path = *pPath;
  u16 x = (x0 + x1) >> 1;
  u16   xInPath;
  u8    yMin, yMax, yInPath;
  bool  isAbobePath, isBelowPath;

  // Update current path pointer
  while (path->x1 <= x) path = ++(*pPath);
  while (path->x0 > x) path = --(*pPath);

  // Check early bbox escape
  if (y0 < y1) {
    yMin = y0; yMax = y1;
  } else {
    yMin = y1; yMax = y0;
  }
  if (path->yMin > yMax || path->yMax < yMin || path->x0 > x1 || path->x1 < x0) {
    collider->onFloor = false;
    return;
  }

  // Apply the path line equation to both points
  isAbobePath = (path_mx(path, x0) + path->b - y0) >= 0;
  isBelowPath = (path_mx(path, x1) + path->b - y1) <= 0;
  if (isAbobePath && isBelowPath) {
    xInPath = x - path->x0;
    yInPath = path->y0 + path_mx(path, xInPath);
    collider->onFloor = true;
    collider->fx = INT2FIXED(path->x0 + xInPath, FX_SCALE);
    collider->fy = INT2FIXED(yInPath, FY_SCALE);
  } else {
    collider->onFloor = false;
  }
}

void collider_check(Collider* collider, u16 x0, u8 y0, u16 x1, u8 y1) {
  collider_checkWithBlocks(collider, x0, y0, x1, y1);
  if (collider->onBlock) {
    return;
  }

  collider_checkWithPaths(collider, &(collider->platform), x0, y0, x1, y1);
  collider->onPlatform = collider->onFloor;
  if (collider->onPlatform) {
    collider->onPath = false;
    return;
  }

  collider_checkWithPaths(collider, &(collider->path), x0, y0, x1, y1);
  collider->onPath = collider->onFloor;
}

void collider_moveOnPath(Collider* collider, Path** pPath, uf16 fx, uf16 fy, f8 fdx) {
  Path* path;
  Path* oldPath;
  uf16  nfx = fx + fdx;
  u16   x = FIXED2INT(nfx, FX_SCALE);
  i16   xInPath;
  u8    yInPath;

  // Update current path pointer
  path = oldPath = *pPath;
  while (path->x1 <= x) path = ++(*pPath);
  while (path->x0 > x) path = --(*pPath);

  // Check being outside current path escape
  xInPath = (i16)x - path->x0;
  if (xInPath < 0 || xInPath >= path->width) {
    collider->onFloor = false;
    return;
  }

  // If we moved to other path, check if there is a gap or not
  yInPath = path->y0 + path_mx(path, xInPath);
  if (path != oldPath) {
    i16   y = FIXED2INT(fy, FY_SCALE);
    i16   d = (i16)yInPath - y;
    if (ABS(d) > COLLIDER_MAX_GAP) {
      collider->onFloor = false;
      return;
    }
  }

  // Update collider position
  collider->onFloor = true;
  collider->fx = nfx + (fdx > 0? path->slopeEffort: -path->slopeEffort);
  // TODO apply effort only when going up (have no sense going down?)
  collider->fy = INT2FIXED((u16)yInPath, FY_SCALE);
}

void collider_moveOnBlock(Collider* collider, uf16 fx, f8 fdx) {
  Block*  block = collider->block;
  uf16    nfx = fx + fdx;
  u16     x = FIXED2INT(nfx, FX_SCALE);

  // Update current block pointer
  while (block->x1 <= x) block = ++collider->block;
  while (block->x0 > x) block = --collider->block;

  // Check being outside current block escape
  if (block->x0 > x || block->x1 <= x) {
    collider->onFloor = collider->onBlock = false;
    return;
  }

  // Update collider position
  collider->onFloor = collider->onBlock = true;
  collider->fx = nfx;
  collider->fy = INT2FIXED(block->y0, FY_SCALE);
}

void collider_move(Collider* collider, uf16 fx, uf16 fy, f8 fdx) {
  if (collider->onBlock) {
    collider_moveOnBlock(collider, fx, fdx);
  } else if (collider->onPlatform) {
    collider_moveOnPath(collider, &(collider->platform), fx, fy, fdx);
    collider->onPlatform = collider->onFloor;
    collider->onPath = false;
  } else if (collider->onPath) {
    collider_moveOnPath(collider, &(collider->path), fx, fy, fdx);
    collider->onPath = collider->onFloor;
    collider->onPlatform = false;
  } else {
    u16 x = FIXED2INT(fx, FX_SCALE);
    u8  y = FIXED2INT(fy, FY_SCALE);
    collider_check(collider, x, y - COLLIDER_MAX_GAP, x, y + COLLIDER_MAX_GAP);
  }
}
