#include "dogs.h"
#include "id.h"
#include "collider.h"
#include "fixed.h"
#include "player.h"
#include "renderer.h"
#include "sprites.h"
#include "chapters.h"
#include "scene.h"
#include "sprite.h"

#define ABS(x)  (x < 0? -x: x)

#define DOG_FX_SCALE   3
#define DOG_FY_SCALE   3

#define DOG_LOOKING_LEFT          0
#define DOG_LOOKING_RIGHT         1
#define DOG_CHANGE_DIRECTION(dog) (1 - dog->direction)

#define DOG_FRONT_CX      12
#define DOG_FRONT_CY      25
#define DOG_FRONT_HEIGHT  10

#define DOG_DYING_DX      6
#define DOG_DYING_DY      10

#define DOG_SLEEPING_CX   8
#define DOG_SLEEPING_CY   DOG_SLEEPING_HEIGHT

#define DOG_STANDING_DISTANCE 50
#define DOG_STANDING_CX       8
#define DOG_STANDING_CY       DOG_STANDING_HEIGHT
#define DOG_STADING_TIME      200

#define DOG_RUNNING_DISTANCE_X  70
#define DOG_RUNNING_DISTANCE_Y  30
#define DOG_RUNNING_VX          16
#define DOG_RUNNING_CX          10
#define DOG_RUNNING_CY          DOG_RUNNING_HEIGHT
#define DOG_RUNNING_RATE        2
#define DOG_RUNNING_FRAME(dog)  (dog->frame++ / DOG_RUNNING_RATE) % DOG_RUNNING_COUNT

#define CLOWN_ACTIVE_DISTANCE       SCREEN_WIDTH / 3
#define CLOWN_FRONT_CX              6
#define CLOWN_FRONT_CY              20
#define CLOWN_FRONT_HEIGHT          15
#define CLOWN_WALKING_VX            3
#define CLOWN_WALKING_CX            8
#define CLOWN_WALKING_CY            CLOWN_WALKING_HEIGHT
#define CLOWN_WALKING_RATE          4
#define CLOWN_WALKING_FRAME(clown)  (clown->frame++ / CLOWN_WALKING_RATE) % CLOWN_WALKING_COUNT
const u8 clown_walkingCX[4] = {0, 0, 4, 0};

#define DOG_FIRST_SPRITE  DOG_SLEEPING_SPRITE_0
#define DOG_SPRITE_COUNT  DOG_SLEEPING_COUNT + DOG_STANDING_COUNT +  \
                          DOG_RUNNING_COUNT + CLOWN_WALKING_COUNT
bool    dog_flipped[DOG_SPRITE_COUNT];

typedef struct {
  // Entity unique id
  u8      id;
  // Integer position
  u16     x;
  u8      y;
  // Fixed point position and velocity
  uf16    fx, fy;
  f8      vx, vy;
  // State and frame inside state
  void    (*step)(void* dog);
  u8      direction;
  u16     frame;
  // Center, size and sprite drawer
  u8      cx, cy;
  u8      width, height;
  u8      sprite;
  // Path colliders
  Collider floorCollider;
  Collider frontCollider;
} Dog;

Dog dogs[SCENE_DOG_COUNT];
u8  dogs_count;

bool dog_isFrontColliding(Dog* dog);

void dog_startSleeping(Dog* dog);
void dog_sleep(void* dog);
void dog_startStanding(Dog* dog);
void dog_stand(void* dog);
void dog_startRunning(Dog* dog);
void dog_run(void* dog);

void clown_startWalking(Dog* clown);
void clown_walk(void* clown);
void clown_startTurning(Dog* clown);

void dogs_init() {
  dogs_count = 0;

  for (u8 i = 0; i < DOG_SPRITE_COUNT; i++) dog_flipped[i] = false;
}

void dogs_reset() {
  dogs_count = 0;

  for (u8 i = 0; i < DOG_SPRITE_COUNT; i++){
    if (dog_flipped[i]) {
      dog_flipped[i] = false;
      sprite_hFlipAsm((u8*)sprites_drawers[i + DOG_FIRST_SPRITE]);
    }
  }
}

void dogs_flipSpriteIfNecessary(Dog* dog) {
  u8 current = dog->sprite;
  u8 index = current - DOG_FIRST_SPRITE;
  if (dog_flipped[index] != dog->direction) {
    dog_flipped[index] = dog->direction;
    sprite_hFlipAsm((u8*)sprites_drawers[current]);
  }
}

Dog* entities_add(u16 x, u8 y) {
  Dog* dog = &dogs[dogs_count++];
  dog->id = id_generate();
  dog->x = x;
  dog->y = y;
  dog->fx = INT2FIXED(x, DOG_FX_SCALE);
  dog->fy = INT2FIXED(y, DOG_FY_SCALE);
  dog->direction = DOG_LOOKING_LEFT;
  collider_create(&(dog->floorCollider));
  collider_create(&(dog->frontCollider));
  return dog;
}

void dogs_add(u16 x, u8 y) {
  Dog* dog = entities_add(x, y);
  dog_startSleeping(dog);
}

void clowns_add(u16 x, u8 y) {
  Dog* clown = entities_add(x, y);
  clown_startWalking(clown);
}

u16 dog_playerDistance(Dog* dog) {
  i16   d = (i16)player_x - (i16)dog->x;
  return ABS(d);
}

void dog_checkForDying(Dog* dog) {
  u8 y0 = dog->y - (dog->cy >> 1);
  if (player_x >= dog->x - dog->cx && player_x <= dog->x + dog->cx &&
      player_y >= y0 && player_y <= dog->y + dog->cx) {
    chapters_reset();
  }
}

void dog_updatePosition(Dog* dog) {
  dog->x = FIXED2INT(dog->fx, DOG_FX_SCALE);
  dog->y = FIXED2INT(dog->fy, DOG_FY_SCALE);
  dog_checkForDying(dog);
}

void clown_startWalking(Dog* clown) {
  clown->vx = clown->direction == DOG_LOOKING_LEFT? -CLOWN_WALKING_VX: CLOWN_WALKING_VX;
  clown->vy = 0;
  clown->step = clown_walk;
  clown->cx = CLOWN_WALKING_CX;
  clown->cy = CLOWN_WALKING_CY;
  clown->width = CLOWN_WALKING_WIDTH;
  clown->height = CLOWN_WALKING_HEIGHT;
  clown->sprite = CLOWN_WALKING_SPRITE_0;
}

bool clown_isActive(Dog* clown) {
  return dog_playerDistance(clown) < CLOWN_ACTIVE_DISTANCE;
}

void clown_walk(void* payload) {
  Dog* clown = (Dog*)payload;

  if (!clown_isActive(clown)) {
    return;
  }

  if (dog_isFrontColliding(clown)) {
    clown_startTurning(clown);
    return;
  }

  collider_move(&(clown->floorCollider), clown->fx, clown->fy, clown->vx);
  if (clown->floorCollider.onFloor) {
    u8  frame = CLOWN_WALKING_FRAME(clown);
    clown->sprite = CLOWN_WALKING_SPRITE_0 + frame;
    clown->cx = clown->direction == DOG_LOOKING_LEFT?
                CLOWN_WALKING_CX + clown_walkingCX[frame]:
                CLOWN_WALKING_CX - clown_walkingCX[frame];
    clown->fx = clown->floorCollider.fx;
    clown->fy = clown->floorCollider.fy;
    dog_updatePosition(clown);
    dogs_flipSpriteIfNecessary(clown);
  } else {
    clown_startTurning(clown);
  }
}

void clown_startTurning(Dog* clown) {
  clown->direction = DOG_CHANGE_DIRECTION(clown);
  clown_startWalking(clown);
  dogs_flipSpriteIfNecessary(clown);
}

void dog_startSleeping(Dog* dog) {
  dog->vx = dog->vy = 0;
  dog->step = dog_sleep;
  dog->cx = DOG_SLEEPING_CX;
  dog->cy = DOG_SLEEPING_CY;
  dog->width = DOG_SLEEPING_WIDTH;
  dog->height = DOG_SLEEPING_HEIGHT;
  dog->sprite = DOG_SLEEPING_SPRITE_0;
}

bool dog_shouldGetStanding(Dog* dog) {
  return dog_playerDistance(dog) < DOG_STANDING_DISTANCE;
}

void dog_sleep(void* payload) {
  Dog*  dog = (Dog*)payload;
  if (dog_shouldGetStanding(dog))
    dog_startStanding(dog);
}

void dog_startStanding(Dog* dog) {
  dog->frame = 0;
  dog->vx = dog->vy = 0;
  dog->step = dog_stand;
  dog->cx = DOG_STANDING_CX;
  dog->cy = DOG_STANDING_CY;
  dog->width = DOG_STANDING_WIDTH;
  dog->height = DOG_STANDING_HEIGHT;
  dog->sprite = DOG_STANDING_SPRITE_0;
  dogs_flipSpriteIfNecessary(dog);
}

bool dog_shouldRun(Dog* dog) {
  return dog_playerDistance(dog) < DOG_RUNNING_DISTANCE_X &&
         (player_y > (dog->y - DOG_RUNNING_DISTANCE_Y / 3) &&
         player_y < (dog->y + DOG_RUNNING_DISTANCE_Y));
}

bool dog_isFrontColliding(Dog* dog) {
  u16 fx = dog->direction == DOG_LOOKING_LEFT? dog->x - DOG_FRONT_CX: dog->x + DOG_FRONT_CX;
  u8  fy = dog->y - DOG_FRONT_CY;
  collider_check(&(dog->frontCollider), fx, fy, fx, fy + DOG_FRONT_HEIGHT);
  return dog->frontCollider.onFloor || dog->frontCollider.onBlock;
}

bool dog_canRun(Dog* dog) {
  collider_move(&(dog->floorCollider), dog->fx, dog->fy, dog->vx);
  return dog->floorCollider.onFloor;
}

void dog_stand(void* payload) {
  Dog* dog = (Dog*)payload;
  dog->frame++;
  if (dog->frame == DOG_STADING_TIME) {
    if (dog_shouldGetStanding(dog)) dog->frame = 0;
    else dog_startSleeping(dog);
  } else if (dog_shouldRun(dog) && dog_canRun(dog) && !dog_isFrontColliding(dog)) {
    dog_startRunning(dog);
  }
}

void dog_startRunning(Dog* dog) {
  dog->direction = player_x > dog->x;
  dog->vx = dog->direction == DOG_LOOKING_LEFT? -DOG_RUNNING_VX: +DOG_RUNNING_VX;
  dog->vy = 0;
  dog->step = dog_run;
  dog->cx = DOG_RUNNING_CX;
  dog->cy = DOG_RUNNING_CY;
  dog->width = DOG_RUNNING_WIDTH;
  dog->height = DOG_RUNNING_HEIGHT;
  dog_run(dog);
}

void dog_run(void* payload) {
  Dog* dog = (Dog*)payload;

  if (dog_isFrontColliding(dog)) {
    dog_startStanding(dog);
    return;
  }

  collider_move(&(dog->floorCollider), dog->fx, dog->fy, dog->vx);
  if (dog->floorCollider.onFloor) {
    dog->sprite = DOG_RUNNING_SPRITE_0 + DOG_RUNNING_FRAME(dog);
    dog->fx = dog->floorCollider.fx;
    dog->fy = dog->floorCollider.fy;
    dog_updatePosition(dog);
    dogs_flipSpriteIfNecessary(dog);
  } else {
    // TODO jump
    dog_startStanding(dog);
  }
}

void dogs_clear() {
  Dog*  dog = dogs;
  u8    count = dogs_count;
  while (count--) {
    if (renderer_isVisible(dog->x - dog->cx) && renderer_isVisible(dog->x + dog->cx)) {
      renderer_clear(dog->id, dog->x - dog->cx, dog->y - dog->cy,
                     dog->width, dog->height);
    }
    dog++;
  }
}

void dog_drawSprite(void* payload, u16 x, u8 y) {
  Dog*  dog = (Dog*)payload;
  u8*   scr = renderer_getScreenPointer(x, y);
  sprites_drawers[dog->sprite](scr);
}

void dogs_render() {
  Dog*  dog = dogs;
  u8    count = dogs_count;
  while (count--) {
    if (renderer_isVisible(dog->x - dog->cx) && renderer_isVisible(dog->x + dog->cx)) {
      bool  isLookingLeft = dog->direction == DOG_LOOKING_LEFT;
      u16   x = isLookingLeft? dog->x: dog->x + dog->width;
      renderer_draw(dog->id,
                    x - dog->cx, dog->y - dog->cy,
                    dog->width, dog->height,
                    &dog_drawSprite, (void*)dog);
    }
    dog++;
  }
}

void dogs_step() {
  Dog*  dog = dogs;
  u8    count = dogs_count;
  while (count--) {
    dog->step(dog);
    dog++;
  }
}
