#include <cpctelera.h>
#include <types.h>
#include "fixed.h"
#include "player.h"
#include "collider.h"
#include "blocks.h"
#include "renderer.h"
#include "sprite.h"
#include "sprites.h"
#include "id.h"
#include "debug.h"
#include "chapters.h"
#include "controls.h"

#define PLAYER_LOOKING_LEFT     0
#define PLAYER_LOOKING_RIGHT    1
#define PLAYER_CHANGE_DIRECTION (1 - player_direction)

#define PLAYER_FX_SCALE   3
#define PLAYER_FY_SCALE   3

#define PLAYER_FRONT_CX       6
#define PLAYER_FRONT_CY       25
#define PLAYER_FRONT_HEIGHT   17

#define PLAYER_START_CLIMBING_CX      6
#define PLAYER_START_CLIMBING_CY      30
#define PLAYER_START_CLIMBING_HEIGHT  20

#define PLAYER_CLIMBING_FY      INT2FIXED(20, PLAYER_FY_SCALE)
#define PLAYER_CLIMBING_CX      6
#define PLAYER_CLIMBING_CY      PLAYER_CLIMBING_HEIGHT
#define PLAYER_CLIMBING_RATE    3
#define PLAYER_CLIMBING_FRAME   (player_frame++ / PLAYER_CLIMBING_RATE)
const u8 player_climbingCX[6] = {0, 4, 6, 6, 6, 6};
const u8 player_climbingCY[6] = {0, 5, 5, 5, 12, 18};

#define PLAYER_PULLING_CX     4
#define PLAYER_PULLING_CY     PLAYER_PULLING_HEIGHT
#define PLAYER_PULLING_WAIT   10
#define PLAYER_PULLING_TIME   16
#define PLAYER_PULLING_DX     2
#define PLAYER_PULLING_FDX    INT2FIXED(PLAYER_PULLING_DX, PLAYER_FX_SCALE)

#define PLAYER_STANDING_CX      2
#define PLAYER_STANDING_CY      PLAYER_STANDING_HEIGHT

#define PLAYER_RUNNING_RATE   2
#define PLAYER_RUNNING_FRAME  (player_frame++ / PLAYER_RUNNING_RATE) % PLAYER_RUNNING_COUNT
#define PLAYER_RUNNING_VX     12
#define PLAYER_RUNNING_CX     6
#define PLAYER_RUNNING_CY     PLAYER_RUNNING_HEIGHT

#define PLAYER_FLYING_CX  4
#define PLAYER_FLYING_CY  PLAYER_FLYING_HEIGHT

#define PLAYER_MIN_FY     INT2FIXED(PLAYER_FLYING_HEIGHT, PLAYER_FY_SCALE)

#define PLAYER_VX_SCALE     FX_SCALE
#define PLAYER_VY_SCALE     FY_SCALE
#define PLAYER_VY_JUMP      -32
#define PLAYER_VY_GRAVITY   4

#define PLAYER_FIRST_SPRITE PLAYER_RUNNING_SPRITE_0
#define PLAYER_SPRITE_COUNT PLAYER_RUNNING_COUNT + PLAYER_STANDING_COUNT +  \
                            PLAYER_JUMPING_COUNT + PLAYER_PULLING_COUNT + \
                            PLAYER_FLYING_COUNT + PLAYER_CLIMBING_COUNT

// Player unique id
u8      player_id;
// Integer position
u16     player_x;
u8      player_y;
// Fixed point position and velocity
uf16    player_fx, player_fy;
f8      player_vx, player_vy;
// State and frame inside state
void    (*player_step)();
u8      player_direction;
u16     player_frame;
// Center, size and current sprite
i8      player_cx, player_cy;
u8      player_width, player_height;
u8      player_sprite;
bool    player_flipped[PLAYER_SPRITE_COUNT];
// Path colliders
Collider player_floorCollider;
Collider player_frontCollider;
// Curring pulling/pushing block
Block*   player_block;

bool player_fallsIntoTheVoid() {
  if (player_x < 296) {
    return player_y >= 160;
  } else if (player_x < 496) {
    return player_y >= 152;
  } else if (player_x < 696) {
    return player_y >= 144;
  } else if (player_x < 896) {
    return player_y >= 136;
  } else if (player_x < 1096) {
    return player_y >= 128;
  } else if (player_x < 1296) {
    return player_y >= 120;
  } else if (player_x < 1496) {
    return player_y >= 112;
  } else if (player_x < 1696) {
    return player_y >= 106;
  } else if (player_x < 1896) {
    return player_y >= 98;
  } else /*if (player_x < 2096)*/ {
    return player_y >= 90;
  }
}

inline void updatePlayerX() {
  player_x = FIXED2INT(player_fx, PLAYER_FX_SCALE);
}
inline void updatePlayerY() {
  player_y = FIXED2INT(player_fy, PLAYER_FY_SCALE);
  if (player_fallsIntoTheVoid()) {
    chapters_reset();
  }
}
void updatePlayerPosition() {
  updatePlayerX();
  updatePlayerY();
}

void startStanding();
void stand();
void turn();
void startRunning();
void run();
void startPulling();
void pull();
void startFlying();
void fly();
void startJumping();
void jump();
void startClimbing();
void climb();

void player_init() {
  player_id = id_generate();

  for (u8 i = 0; i < PLAYER_SPRITE_COUNT; i++) player_flipped[i] = false;
}

void player_reset(u16 x, u8 y) {
  player_fx = INT2FIXED(x, PLAYER_FX_SCALE);
  player_fy = INT2FIXED(y, PLAYER_FY_SCALE);
  player_vy = player_vx = 0;
  updatePlayerPosition();

  player_direction = PLAYER_LOOKING_LEFT;

  collider_create(&player_floorCollider);
  collider_create(&player_frontCollider);

  for (u8 i = 0; i < PLAYER_SPRITE_COUNT; i++) {
    if (player_flipped[i]) {
      player_flipped[i] = false;
      sprite_hFlipAsm((u8*)sprites_drawers[i + PLAYER_FIRST_SPRITE]);
    }
  }

  startFlying();
}

void flipSpriteIfNecessary() {
  u8 index = player_sprite - PLAYER_FIRST_SPRITE;
  if (player_flipped[index] != player_direction) {
    player_flipped[index] = player_direction;
    sprite_hFlipAsm((u8*)sprites_drawers[player_sprite]);
  }
}

void startStanding() {
  player_step = stand;
  player_vx = player_vy = 0;
  player_cx = PLAYER_STANDING_CX;
  player_cy = PLAYER_STANDING_CY;
  player_width = PLAYER_STANDING_WIDTH;
  player_height = PLAYER_STANDING_HEIGHT;
  player_sprite = PLAYER_STANDING_SPRITE_0;
}

bool isTryingToClimb() {
  if (controls_keyUp()) {
    u16 fx = player_direction == PLAYER_LOOKING_LEFT?
             player_x - PLAYER_START_CLIMBING_CX:
             player_x + PLAYER_START_CLIMBING_CX;
    u8  fy = player_y - PLAYER_START_CLIMBING_CY;
    collider_check(&player_frontCollider, fx, fy, fx, fy + PLAYER_START_CLIMBING_HEIGHT);
    if (player_frontCollider.onFloor) {
      return true;
    }
  }
  return false;
}

bool canPull() {
  u16 fx = player_direction == PLAYER_LOOKING_LEFT?
           player_x - PLAYER_START_CLIMBING_CX:
           player_x + PLAYER_START_CLIMBING_CX;
  u8  fy = player_y - PLAYER_START_CLIMBING_CY;
  collider_check(&player_frontCollider, fx, fy, fx, fy + PLAYER_START_CLIMBING_HEIGHT);
  return player_frontCollider.onBlock && player_frontCollider.block->climbable;
}

void stand() {
  bool isLookingLeft = (player_direction == PLAYER_LOOKING_LEFT);

  if (isTryingToClimb()) {
    startClimbing();
    return;
  }

  if (controls_keyLeft())
    if (isLookingLeft) startRunning(); else turn();
  else if (controls_keyRight())
    if (isLookingLeft) turn(); else startRunning();
  else if (controls_keyAction())
    if (canPull()) startPulling();

  flipSpriteIfNecessary();
}

void turn() {
  player_direction = PLAYER_CHANGE_DIRECTION;
  updatePlayerX();
  flipSpriteIfNecessary();
}

void startRunning() {
  player_step = run;
  player_vx = player_direction == PLAYER_LOOKING_LEFT? -PLAYER_RUNNING_VX: +PLAYER_RUNNING_VX;
  player_frame = 0;
  player_cx = PLAYER_RUNNING_CX;
  player_cy = PLAYER_RUNNING_CY;
  player_width = PLAYER_RUNNING_WIDTH;
  player_height = PLAYER_RUNNING_HEIGHT;
  run();
}


bool isCollidingFrontOrBack(u16 fx) {
  u8  fy = player_y - PLAYER_FRONT_CY;
  collider_check(&player_frontCollider, fx, fy, fx, fy + PLAYER_FRONT_HEIGHT);
  return player_frontCollider.onFloor || player_frontCollider.onBlock;
}
bool isFrontColliding() {
  u16 fx = player_direction == PLAYER_LOOKING_LEFT? player_x - PLAYER_FRONT_CX: player_x + PLAYER_FRONT_CX;
  return isCollidingFrontOrBack(fx);
}
bool isBackColliding() {
  u16 fx = player_direction == PLAYER_LOOKING_LEFT? player_x + PLAYER_FRONT_CX: player_x - PLAYER_FRONT_CX;
  return isCollidingFrontOrBack(fx);
}

void run() {
  bool isLookingLeft = player_direction == PLAYER_LOOKING_LEFT;

  if ((controls_keyLeft() && isLookingLeft) ||
      (controls_keyRight() && !isLookingLeft)) {
    if (controls_keyUp()) {
      startJumping();
      return;
    }
  } else {
    startStanding();
    return;
  }

  if (isFrontColliding()) {
    startStanding();
    return;
  }

  collider_move(&player_floorCollider, player_fx, player_fy, player_vx);
  if (player_floorCollider.onFloor) {
    player_sprite = PLAYER_RUNNING_SPRITE_0 + PLAYER_RUNNING_FRAME;
    player_vy = (f16)player_floorCollider.fy - (f16)player_fy;
    player_fx = player_floorCollider.fx;
    player_fy = player_floorCollider.fy;
    updatePlayerPosition();
    flipSpriteIfNecessary();
  } else {
    player_fx += player_vx;
    startFlying();
  }
}

void startFlying() {
  player_step = fly;
  player_cx = PLAYER_FLYING_CX;
  player_cy = PLAYER_FLYING_CY;
  player_width = PLAYER_FLYING_WIDTH;
  player_height = PLAYER_FLYING_HEIGHT;
  player_sprite = PLAYER_FLYING_SPRITE_0;
  fly();
}

void fly() {
  uf16  nfx, nfy;
  u16   nx;
  u8    ny;

  if (isTryingToClimb()) {
    startClimbing();
    return;
  }

  if (isFrontColliding()) {
    player_vx = 0;
  }

  player_vy += PLAYER_VY_GRAVITY;

  nfx = player_fx + player_vx;
  nfy = player_fy + player_vy;
  if (nfy < PLAYER_MIN_FY)
    nfy = PLAYER_MIN_FY;
  nx = FIXED2INT(nfx, PLAYER_FX_SCALE);
  ny = FIXED2INT(nfy, PLAYER_FY_SCALE);
  collider_check(&player_floorCollider, player_x, player_y, nx, ny);
  if (player_floorCollider.onFloor) {
    player_fx = player_floorCollider.fx;
    player_fy = player_floorCollider.fy;
    startStanding();
  } else {
    player_fx = nfx;
    player_fy = nfy;
    flipSpriteIfNecessary();
  }

  updatePlayerPosition();
}

void startJumping() {
  player_frame = 0;
  player_step = jump;
  jump();
}

void jump() {
  u8 frame = PLAYER_RUNNING_FRAME;
  if (frame < PLAYER_JUMPING_COUNT) {
    if (isFrontColliding()) {
      startStanding();
      return;
    }
    player_sprite = PLAYER_JUMPING_SPRITE_0 + frame;
    player_fx += player_vx;
    player_fy += player_vy;
    updatePlayerPosition();
    flipSpriteIfNecessary();
  } else {
    player_vy = PLAYER_VY_JUMP;
    startFlying();
  }
}

void startPulling() {
  player_step = pull;
  player_cx = PLAYER_PULLING_CX;
  player_cy = PLAYER_PULLING_CY;
  player_width = PLAYER_PULLING_WIDTH;
  player_height = PLAYER_PULLING_HEIGHT;
  player_sprite = PLAYER_PULLING_SPRITE_0;
  player_frame = 0;
  player_block = player_frontCollider.block;
  pull();
}

void pull() {
  bool isLookingLeft = player_direction == PLAYER_LOOKING_LEFT;

  if (!controls_keyAction() || isBackColliding()) {
    startStanding();
    return;
  }

  if ((controls_keyLeft() && !isLookingLeft) ||
      (controls_keyRight() && isLookingLeft)) {
    if (player_frame == 0) {
      i16 dx = isLookingLeft? PLAYER_PULLING_DX: -PLAYER_PULLING_DX;
      blocks_clear(player_block);
      player_block->x0 += dx;
      player_block->x1 += dx;

      player_sprite = PLAYER_PULLING_SPRITE_1;
      player_cx = isLookingLeft? PLAYER_PULLING_CX - 2: PLAYER_PULLING_CX + 2;
      player_frame++;
    } else if (player_frame == PLAYER_PULLING_WAIT) {
      f16 dx = isLookingLeft? PLAYER_PULLING_FDX: -PLAYER_PULLING_FDX;
      player_sprite = PLAYER_PULLING_SPRITE_0;
      player_cx = PLAYER_PULLING_CX;
      collider_move(&player_floorCollider, player_fx, player_fy, dx);
      if (player_floorCollider.onFloor) {
        player_fx = player_floorCollider.fx;
        player_fy = player_floorCollider.fy;
        renderer_forceRedraw(player_id);
        updatePlayerPosition();
        player_frame++;
      } else {
        player_fx += dx;
        startFlying();
        return;
      }
    } else if (player_frame == PLAYER_PULLING_TIME) {
      player_frame = 0;
    } else {
      player_frame++;
    }
    flipSpriteIfNecessary();
  }
}

void startClimbing() {
  player_step = climb;
  player_frame = 0;
  player_vx = player_vy = 0;
  player_width = PLAYER_CLIMBING_WIDTH;
  player_height = PLAYER_CLIMBING_HEIGHT;
  player_fy = player_frontCollider.fy + PLAYER_CLIMBING_FY;
  updatePlayerPosition();
  climb();
}

void climb() {
  u8 frame = PLAYER_CLIMBING_FRAME;
  if (frame < PLAYER_CLIMBING_COUNT) {
    player_sprite = PLAYER_CLIMBING_SPRITE_0 + frame;
    player_cx = player_direction == PLAYER_LOOKING_LEFT?
                PLAYER_CLIMBING_CX + player_climbingCX[frame]:
                PLAYER_CLIMBING_CX - player_climbingCX[frame];
    player_cy = PLAYER_CLIMBING_CY + player_climbingCY[frame];
    flipSpriteIfNecessary();
  } else {
    i8 dx = player_direction == PLAYER_LOOKING_LEFT? -6: +6;
    player_fx = INT2FIXED(((i16)player_x + dx), PLAYER_FX_SCALE);
    player_fy = player_frontCollider.fy;
    updatePlayerPosition();

    player_floorCollider.block = player_frontCollider.block;
    player_floorCollider.onBlock = player_frontCollider.onBlock;

    player_floorCollider.platform = player_frontCollider.platform;
    player_floorCollider.onPlatform = player_frontCollider.onPlatform;

    player_floorCollider.path = player_frontCollider.path;
    player_floorCollider.onPath = player_frontCollider.onPath;

    startStanding();
  }
}

void player_clear() {
  renderer_clear(player_id, player_x - player_cx, player_y - player_cy,
                 player_width, player_height);
}

void _drawSprite(void* payload, u16 x, u8 y) {
  u8* scr = renderer_getScreenPointer(x, y);
  sprites_drawers[player_sprite](scr);
}

void player_render() {
  bool isLookingLeft = player_direction == PLAYER_LOOKING_LEFT;
  u16 x = isLookingLeft? player_x: player_x + player_width;
  renderer_draw(player_id,
                x - player_cx, player_y - player_cy,
                player_width, player_height,
                &_drawSprite, (void*)0);
}

bool player_isVisible() {
  return renderer_isVisible(player_x);
}
