#include <cpctelera.h>
#include "renderer.h"
#include "scene.h"
#include "crtc.h"
#include "bool.h"
#include "debug.h"

#define WIDTH_IN_SPANS  SCREEN_WIDTH / SCENE_SPAN_WIDTH

#define INTERRUPT_HALF_SCREEN (3 + 6)
#define MAX_DIRTY_BOXES       8
#define MAX_DRAWERS           8

extern void renderer_drawSpanAsmFast(u8* scr, u8* span) __z88dk_callee;
extern void renderer_drawPartialSpanAsmFast(u8* scr, u8* span, u8 y, u8 height) __z88dk_callee;

u8* renderer_screen;
u16 renderer_scroll;
u16 renderer_fineScroll;

bool  renderer_drawNewSpan;
u8    renderer_newSpanXInBytes;
u16   renderer_newSpanIndex;

u8    renderer_interrupt;

typedef struct {
  u8    id;
  bool  refresh;
  u16   x;
  u8    y;
  u8    width;
  u8    height;
} Box;

typedef struct {
  u8    id;
  bool  refresh;
  u16   x;
  u8    y, width, height;
  void  (*f)(void* payload, u16 x, u8 y);
  void* payload;
} Drawer;

Box dirtyTopHalfBoxes[MAX_DIRTY_BOXES];
u8  dirtyTopHalfBoxCount;
Box dirtyBottomHalfBoxes[MAX_DIRTY_BOXES];
u8  dirtyBottomHalfBoxCount;

Drawer drawers[MAX_DRAWERS];
u8 drawerCount;

// u8* getScreenPointer(u8 xInBytes, u8 y) {
//   return renderer_screen + SCREEN_WITH_IN_BYTES * (y >> 3) + 2048 * (y % 8) + xInBytes;
// }

void myInterruptHandler() {
  // const static u8 colors[12] = {HW_RED, HW_GREEN, HW_BLUE, HW_BRIGHT_RED, HW_BRIGHT_GREEN, HW_BRIGHT_BLUE,
  //                               HW_RED, HW_GREEN, HW_BLUE, HW_BRIGHT_RED, HW_BRIGHT_GREEN, HW_BRIGHT_BLUE};
  renderer_interrupt = (renderer_interrupt + 1) % 12;
  // cpct_setBorder(colors[renderer_interrupt]);
  // cpct_setPALColour(1, colors[renderer_interrupt]);
}

void renderer_init() {
  cpct_setVideoMode(0);
  cpct_setBorder(HW_BLACK);
  for (u8 i = 0; i < 16; i++) {
    cpct_setPALColour(i, HW_BLACK);
  }
  setCRTC();

  renderer_scroll = renderer_fineScroll = 0;
  renderer_screen = CPCT_VMEM_START;

  renderer_interrupt = 5;
  cpct_waitVSYNC();
  while(cpct_count2VSYNC() == 1);
  cpct_setInterruptHandler(myInterruptHandler);
}

void renderer_reset() {
  u8 frame = cpct_px2byteM0(2, 2);
  cpct_memset(CPCT_VMEM_START, frame, 0x4000);
  renderer_begin();
}

void renderer_begin() {
  dirtyTopHalfBoxCount = dirtyBottomHalfBoxCount = drawerCount = 0;
}

void addDirtyBox(Box *list, u8 index, u8 id, u16 x, u8 y, u8 width, u8 height) {
  Box*  box = &list[index];
  bool  modified;

  // Adjust x to tile, preventing non detected overlaps
  u8    rest = x % 4;
  x -= rest;
  width += rest;

  modified = (box->x != x) || (box->y != y) || (box->id != id) ||
             (box->width != width) || (box->height != height);
  if ((box->refresh = modified)) {
    box->id = id;
    box->x = x;
    box->y = y;
    box->width = width;
    box->height = height;
  }
}

void renderer_clear(u8 id, u16 x, u8 y, u8 width, u8 height) {
  u8 y1 = y + height;
  if (y1 <= SCENE_SPAN_HEIGHT) {
    addDirtyBox(dirtyTopHalfBoxes, dirtyTopHalfBoxCount++, id, x, y, width, height);
  } else if (y >= SCENE_SPAN_HEIGHT) {
    addDirtyBox(dirtyBottomHalfBoxes, dirtyBottomHalfBoxCount++, id, x, y, width, height);
  } else {
    u8 topHalfHeight = SCENE_SPAN_HEIGHT - y;
    u8 bottomHalfHeight = y1 - SCENE_SPAN_HEIGHT;
    addDirtyBox(dirtyTopHalfBoxes, dirtyTopHalfBoxCount++, id, x, y, width, topHalfHeight);
    addDirtyBox(dirtyBottomHalfBoxes, dirtyBottomHalfBoxCount++, id, x, SCENE_SPAN_HEIGHT, width, bottomHalfHeight);
  }
}

void renderer_drawSpans(u8* scr, u16 *indices) {
  for (u8 i = 0; i < WIDTH_IN_SPANS; i++) {
    u16 index = *(indices++);
    u8  *span = &scene_spans[index];
    renderer_drawSpanAsmFast(scr, span);
    scr += SCENE_SPAN_WIDTH_IN_BYTES;
  }
}

void renderer_clearAll() {
  u8*   scr;

  scr = getScreenPointerAsm(0, 0);
  renderer_drawSpans(scr, &scene_topHalfIndices[renderer_scroll]);

  scr = getScreenPointerAsm(0, SCENE_SPAN_HEIGHT);
  renderer_drawSpans(scr, &scene_bottomHalfIndices[renderer_scroll]);
}

void renderer_center(u16 x) {
  if (x < SCREEN_WIDTH / 2) {
    renderer_fineScroll = 0;
  } else if (x > SCENE_SPAN_COUNT * SCENE_SPAN_WIDTH - SCREEN_WIDTH / 2) {
    renderer_fineScroll = SCENE_SPAN_COUNT * SCENE_SPAN_WIDTH_IN_BYTES - SCREEN_WIDTH / 2;
  } else {
    renderer_fineScroll = (x - SCREEN_WIDTH / 2) >> 1;
  }
  renderer_scroll = renderer_fineScroll >> 1;
  renderer_screen = CPCT_VMEM_START + (renderer_fineScroll & 0xfffe);
  updateCRTC();
}

void renderer_updateScroll(i8 movement) {
  u16 newScroll;

  renderer_fineScroll += movement;
  newScroll = renderer_fineScroll >> 1;
  if (renderer_scroll != newScroll) {
    renderer_scroll = newScroll;
    renderer_drawNewSpan = true;
    renderer_screen = CPCT_VMEM_START + (renderer_scroll << 1);

    if (movement > 0) {
      renderer_newSpanXInBytes = SCREEN_WITH_IN_BYTES - 2;
      renderer_newSpanIndex = 49 + renderer_scroll;
    } else {
      renderer_newSpanXInBytes = 0;
      renderer_newSpanIndex = 0 + renderer_scroll;
    }
  } else {
    renderer_drawNewSpan = false;
  }
}

void forceDirtyBoxesToRefresh(u8 id);

void renderer_forceRedraw(u8 id) {
  forceDirtyBoxesToRefresh(id);
}

void renderer_draw(u8 id, u16 x, u8 y, u8 width, u8 height,
                   void (*f)(void*, u16, u8), void* payload) {
  Drawer* drawer = &drawers[drawerCount++];
  u8      y1 = y + height;
  bool    modified = (drawer->x != x) || (drawer->y != y) || (drawer->id != id) ||
                     (drawer->width != width) || (drawer->height != height) ||
                     (drawer->f != f) || (drawer->payload != payload);
  if ((drawer->refresh = modified)) {
    drawer->id = id;
    drawer->x = x;
    drawer->y = y;
    drawer->width = width;
    drawer->height = height;
    drawer->f = f;
    drawer->payload = payload;
  }
}

void cleanHalfBox(u16 wx, u8 wy, u8 width, u8 height, u8 wyHalf, u16* indices) {
  // TODO: move to ASM
  u16  index = wx >> 2;
  u8   xInBytes = (index << 1) - (renderer_scroll << 1);
  u8   widthInSpans = ((wx + width) >> 2) - index + 1;
  u16* idx = &indices[index];
  u8*  scr = getScreenPointerAsm(xInBytes, wy);
  while (widthInSpans--) {
    u8* span = &scene_spans[*(idx++)];
    renderer_drawPartialSpanAsmFast(scr, span, wyHalf, height);
    scr += 2;
  }
}

void updateCRTC() {
  if (renderer_fineScroll & 1) {
    setR30();
  } else {
    setR31();
  }
  cpct_setVideoMemoryPage(cpct_pageC0 + (renderer_scroll >> 8));
  cpct_setVideoMemoryOffset(renderer_scroll);
}

void forceDirtyBoxesToRefresh(u8 id) {
  for (u8 i = 0; i < dirtyTopHalfBoxCount; i++) {
    Box* box = &dirtyTopHalfBoxes[i];
    if (box->id == id) {
      box->refresh = true;
      break;
    }
  }
  for (u8 i = 0; i < dirtyBottomHalfBoxCount; i++) {
    Box* box = &dirtyBottomHalfBoxes[i];
    if (box->id == id) {
      box->refresh = true;
      break;
    }
  }
}

bool isDrawerOverlappingBox(Drawer* drawer, Box* box) {
  return (drawer->x + drawer->width) > box->x &&
         (box->x + box->width) >= drawer->x &&
         (drawer->y + drawer->height) > box->y &&
         (box->y + box->height) >= drawer->y;
}

void forceDrawersToRefresh(Box* box) {
  u8 id = box->id;
  Drawer* drawer = drawers;
  for (u8 i = 0; i < drawerCount; i++, drawer++) {
    if (!drawer->refresh && (drawer->id == id || isDrawerOverlappingBox(drawer, box))) {
      drawer->refresh = true;
    }
  }
}

void renderer_end() {
  // Force modified drawers to refresh their dirty boxes
  for (u8 i = 0; i < drawerCount; i++) {
    Drawer* drawer = &drawers[i];
    if (drawer->refresh) {
      forceDirtyBoxesToRefresh(drawer->id);
    }
  }

  for (u8 i = 0; i < dirtyTopHalfBoxCount; i++) {
    Box* box = &dirtyTopHalfBoxes[i];
    if (box->refresh) {
      forceDrawersToRefresh(box);
    }
  }

  for (u8 i = 0; i < dirtyBottomHalfBoxCount; i++) {
    Box* box = &dirtyBottomHalfBoxes[i];
    if (box->refresh) {
      forceDrawersToRefresh(box);
    }
  }

  // Wait for half screen to draw top half
  while (renderer_interrupt != INTERRUPT_HALF_SCREEN);

  debug_setBorder(HW_BRIGHT_BLUE);
  cpct_akp_musicPlay();

  debug_setBorder(HW_GREEN);

  // Clear top half screen
  for (u8 i = 0; i < dirtyTopHalfBoxCount; i++) {
    Box* box = &dirtyTopHalfBoxes[i];
    if (box->refresh) {
      cleanHalfBox(box->x, box->y, box->width, box->height, box->y, scene_topHalfIndices);
      // forceDrawersToRefresh(box);
    }
  }

  // Draw top half new span
  if (renderer_drawNewSpan) {
    u8* scr = getScreenPointerAsm(renderer_newSpanXInBytes, 0);
    u16 index = scene_topHalfIndices[renderer_newSpanIndex];
    u8* span = &scene_spans[index];
    renderer_drawSpanAsmFast(scr, span);
  }

  // Draw top half entities
  for (u8 i = 0; i < drawerCount; i++) {
    Drawer* drawer = &drawers[i];
    if (drawer->refresh && (drawer->y + drawer->height) < SCENE_SPAN_HEIGHT) {
      drawer->f(drawer->payload, drawer->x, drawer->y);
    }
  }

  debug_resetBorder();

  // Wait for end screen to write bottom half
  while (renderer_interrupt == INTERRUPT_HALF_SCREEN || renderer_interrupt == INTERRUPT_HALF_SCREEN + 1);

  debug_setBorder(HW_BLUE);

  // Now, we can move the screen
  updateCRTC();

  // Clear bottom half screen
  for (u8 i = 0; i < dirtyBottomHalfBoxCount; i++) {
    Box* box = &dirtyBottomHalfBoxes[i];
    if (box->refresh) {
      cleanHalfBox(box->x, box->y, box->width, box->height, box->y-SCENE_SPAN_HEIGHT, scene_bottomHalfIndices);
      // forceDrawersToRefresh(box);
    }
  }

  // Draw bottom half new span
  if (renderer_drawNewSpan) {
    u8* scr = getScreenPointerAsm(renderer_newSpanXInBytes, SCENE_SPAN_HEIGHT);
    u16 index = scene_bottomHalfIndices[renderer_newSpanIndex];
    u8* span = &scene_spans[index];
    renderer_drawSpanAsmFast(scr, span);
  }

  // Render bottom half entities
  for (u8 i = 0; i < drawerCount; i++) {
    Drawer* drawer = &drawers[i];
    if (drawer->refresh && (drawer->y + drawer->height) >= SCENE_SPAN_HEIGHT) {
      drawer->f(drawer->payload, drawer->x, drawer->y);
    }
  }

  debug_resetBorder();
}

u8* renderer_getScreenPointer(u16 wx, u8 wy) {
  u8 xInBytes = (wx >> 1) - (renderer_scroll << 1);
  return getScreenPointerAsm(xInBytes, wy);
}

u8 renderer_isVisible(u16 wx) {
  u16 xInChars = (wx >> 2);
  if (xInChars >= renderer_scroll && xInChars < renderer_scroll + SCREEN_WITH_IN_CHARS)
    return 1;
  else
    return 0;
}
